/*
 *      modem-manager-conky.h
 *      
 *      Copyright 2013 Alex <alex@linuxonly.ru>
 *      
 *      This program is free software: you can redistribute it and/or modify
 *      it under the terms of the GNU General Public License as published by
 *      the Free Software Foundation; either version 3 of the License, or
 *      (at your option) any later version.
 *      
 *      This program is distributed in the hope that it will be useful,
 *      but WITHOUT ANY WARRANTY; without even the implied warranty of
 *      MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *      GNU General Public License for more details.
 *      
 *      You should have received a copy of the GNU General Public License
 *      along with this program. If not, see <http://www.gnu.org/licenses/>.
 */

#include <stdlib.h>
#include <string.h>
#include <glib.h>
#include <gio/gio.h>
#include <glib/gprintf.h>

struct _mmc_device {
	gchar *objectpath;
	gchar *manufacturer;
	gchar *modem;
	gchar *version;
};

typedef struct _mmc_device *mmc_device_t;

static gint floatoutput = 0;
static gchar *forcedevice = NULL; 

static GOptionEntry entries[] =
{
  { "floatoutput", 'f', 0, G_OPTION_ARG_INT, &floatoutput, "Output signal level as floating point value", NULL },
  { "forcedevice", 'd', 0, G_OPTION_ARG_STRING, &forcedevice, "Use only specified device (manufacturer modem version)", NULL },
  { NULL }
};


static void modem_manager_conky_error_message(GError *error)
{
	if (error == NULL) return;
	
	g_debug("Error: %s", error->message ? error->message : "Unknown");
	g_error_free(error);
}

static void modem_manager_conky_get_device_status(GDBusConnection *connection, const gchar *devpath, GSList **devices)
{
	GDBusProxy *modemproxy;
	GError *error;
	GVariant *devinfo;
	mmc_device_t device;
	
	if ((connection == NULL) || (devpath == NULL) || (devices == NULL)) return;
	
	error = NULL;
	
	modemproxy = g_dbus_proxy_new_sync(connection,
									G_DBUS_PROXY_FLAGS_NONE,
									NULL,
									"org.freedesktop.ModemManager",
									devpath,
									"org.freedesktop.ModemManager.Modem",
									NULL,
									&error);
	
	if (error != NULL) {
		modem_manager_conky_error_message(error);
		return;
	}
	
	devinfo = g_dbus_proxy_call_sync(modemproxy,
										"GetInfo",
										NULL,
										0,
										-1,
										NULL,
										&error);
	
	if (error != NULL) {
		modem_manager_conky_error_message(error);
		return;
	}
	
	device = (mmc_device_t)g_new0(struct _mmc_device, 1);
	
	g_variant_get(devinfo, "((sss))", &device->manufacturer, &device->modem, &device->version);
	
	device->objectpath = g_strdup(devpath);
	device->manufacturer = g_strdup(device->manufacturer);
	device->modem = g_strdup(device->modem);
	device->version = g_strdup(device->version);
	
	g_variant_unref(devinfo);
	
	g_object_unref(modemproxy);
	
	(*devices) = g_slist_prepend((*devices), device);
}

static void modem_manager_conky_free_device_status(gpointer data, gpointer user_data)
{
	mmc_device_t device;
	
	if (data == NULL) return;
	
	device = (mmc_device_t)data;
	
	if (device->objectpath != NULL) g_free(device->objectpath);
	if (device->manufacturer != NULL) g_free(device->manufacturer);
	if (device->modem != NULL) g_free(device->modem);
	if (device->version != NULL) g_free(device->version);
}

static void modem_manager_conky_free_device_list(GSList *devices)
{
	if (devices == NULL) return;
	
	g_slist_foreach(devices, modem_manager_conky_free_device_status, NULL);
	g_slist_free(devices);
}

static gint modem_manager_conky_search_device_list(gconstpointer a, gconstpointer b)
{
	mmc_device_t device;
	gchar devstring[256];
	gchar *userstring;
	
	device = (mmc_device_t)a;
	userstring = (gchar *)b;
	
	memset(devstring, 0, sizeof(devstring));
	g_snprintf(devstring, sizeof(devstring), "%s %s %s", device->manufacturer, device->modem, device->version);
	
	if (!g_str_equal(devstring, userstring)) {
		return 1;
	} else {
		return 0;
	}
}

static void modem_manager_conky_return_signal_level(gint level, gboolean fp)
{
	if (!fp) {
		g_printf("%u\n", level);
	} else {
		g_printf("%3.2f\n", level/100.0);
	}
}

int main(int argc, char **argv)
{
	GOptionContext *context;
	GDBusConnection *connection;
	GError *error;
	GDBusProxy *managerproxy;
	GVariant *devicesvar;
	GVariantIter deviter, deviter2;
	GVariant *devnode, *devnode2;
	GVariant *devvalue;
	gsize strsize;
	const gchar *devpath;
	GSList *devices;
	GSList *devicefound;
	mmc_device_t device;
	GDBusProxy *netproxy;
	GVariant *sigvalue;
	guint signal;
	
	#if !GLIB_CHECK_VERSION (2, 36, 0)
		g_type_init();
	#endif
	
	error = NULL;
	
	context = g_option_context_new("- signal level information bridge between ModemManager and Conky");
	g_option_context_add_main_entries(context, entries, NULL);
	if (!g_option_context_parse(context, &argc, &argv, &error)) {
		g_printf("Option parsing failed: %s\n", error->message);
		g_error_free(error);
		return 1;
	}
	
	error = NULL;
	
	connection = g_bus_get_sync(G_BUS_TYPE_SYSTEM, NULL, NULL);
	
	if (error != NULL) {
		modem_manager_conky_error_message(error);
		modem_manager_conky_return_signal_level(0, (gboolean)floatoutput);
		return 1;
	}
	
	error = NULL;
	
	managerproxy = g_dbus_proxy_new_sync(connection,
										G_DBUS_PROXY_FLAGS_NONE,
										NULL,
										"org.freedesktop.ModemManager",
										"/org/freedesktop/ModemManager",
										"org.freedesktop.ModemManager",
										NULL,
										&error);
	
	if (error != NULL) {
		modem_manager_conky_error_message(error);
		g_object_unref(connection);
		modem_manager_conky_return_signal_level(0, (gboolean)floatoutput);
		return 1;
	}
	
	devicesvar = g_dbus_proxy_call_sync(managerproxy,
										"EnumerateDevices",
										NULL,
										0,
										-1,
										NULL,
										&error);
										
	if (error != NULL) {
		modem_manager_conky_error_message(error);
		g_object_unref(managerproxy);
		g_object_unref(connection);
		modem_manager_conky_return_signal_level(0, (gboolean)floatoutput);
		return 1;
	}
	
	devices = NULL;
	
	g_variant_iter_init(&deviter, devicesvar);
	
	while ((devnode = g_variant_iter_next_value(&deviter)) != NULL) {
		g_variant_iter_init(&deviter2, devnode);
		while ((devnode2 = g_variant_iter_next_value(&deviter2)) != NULL) {
			strsize = 256;
			devpath = g_variant_get_string(devnode2, &strsize);
			modem_manager_conky_get_device_status(connection, devpath, &devices);
			g_variant_unref(devnode2);
		}
		g_variant_unref(devnode);
    }
	
	g_object_unref(managerproxy);
	
	if (g_slist_length(devices) == 0) {
		g_object_unref(connection);
		modem_manager_conky_return_signal_level(0, (gboolean)floatoutput);
		return 1;
	}
	
	devices = g_slist_reverse(devices);
	
	if (forcedevice != NULL) {
		devicefound = g_slist_find_custom(devices, forcedevice, modem_manager_conky_search_device_list);
		if (devicefound == NULL) {
			g_object_unref(connection);
			modem_manager_conky_return_signal_level(0, (gboolean)floatoutput);
			return 1;
		} else {
			device = devicefound->data;
		}
	} else {
		device = g_slist_nth_data(devices, 0);
	}
	
	error = NULL;
	
	netproxy = g_dbus_proxy_new_sync(connection,
										G_DBUS_PROXY_FLAGS_NONE,
										NULL,
										"org.freedesktop.ModemManager",
										device->objectpath,
										"org.freedesktop.ModemManager.Modem.Gsm.Network",
										NULL,
										&error);
	
	if ((error != NULL) && (netproxy == NULL)) {
		modem_manager_conky_error_message(error);
		modem_manager_conky_free_device_list(devices);
		g_object_unref(connection);
		modem_manager_conky_return_signal_level(0, (gboolean)floatoutput);
		return 1;
	}
	
	sigvalue = g_dbus_proxy_call_sync(netproxy,
										"GetSignalQuality",
										NULL,
										0,
										-1,
										NULL,
										&error);
	
	if ((error != NULL) && (sigvalue == NULL)) {
		modem_manager_conky_error_message(error);
		modem_manager_conky_free_device_list(devices);
		g_object_unref(netproxy);
		g_object_unref(connection);
		modem_manager_conky_return_signal_level(0, (gboolean)floatoutput);
		return 1;
	}
	
	g_variant_get(sigvalue, "(u)", &signal);
	
	g_variant_unref(sigvalue);
	
	modem_manager_conky_free_device_list(devices);
	
	g_object_unref(netproxy);
	g_object_unref(connection);
	
	modem_manager_conky_return_signal_level(signal, (gboolean)floatoutput);
	
	return 0;
}
